#include <sys/ioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <qdirectpainter_qws.h>
#include <qdir.h>
#include <qevent.h>
#include <qpe/qpeapplication.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <qmenubar.h>

#include "version.h"
#include "pmiroir.h"
#include "pmimage.h"
#include "filedialog.h"
#include "sketchpainter.h"

PMiroir::PMiroir(QWidget *parent) : QMainWindow(parent, 0) {
  setup_menu();

  _timer = new QTimer(this);
  _mode = MirrorMode;
  _fullrect.setCoords(0, 0, Width, Height);
  _imagedir = QString();

  _image = new PMImage(this, _imagedir);
  _painter = new SketchPainter(Width, Height);
  _pen1 = QPen(red, 5);
  _pen2 = QPen(yellow, 20);
  _pen3 = QPen(white, 20);
  _pen = &_pen1;

  _buffer0 = new unsigned short int[Width * Height];
  _buffer1 = new unsigned short int[Width * Height];
  _check0 = new unsigned long int[Height];
  _check1 = new unsigned long int[Height];

  _buf = _buffer0;
  _bbuf = _buffer1;
  _chk = _check0;

  _vgafp = open(CFXGADev, O_RDWR);
  //  if (_vgafp == -1) exit(0);
  ioctl(_vgafp, CFXGA_SET_RESO, 0);
  ioctl(_vgafp, CFXGA_CLEAR_SCREEN, NULL);

  connect(_image, SIGNAL(fileLoaded(QString)),
          _painter, SLOT(loadOverLayer(QString)));
  connect(_image, SIGNAL(fileClosed(QString)),
          _painter, SLOT(saveOverLayer(QString)));

  connect(_timer, SIGNAL(timeout()), this, SLOT(reload_buf()));
  _timer->start(500, true);
}


PMiroir::~PMiroir(void) {
  delete _image;
  delete _painter;
  delete _buffer0;
  delete _buffer1;
  delete _check0;
  delete _check1;
  delete _timer;
  close(_vgafp);
}

void PMiroir::quit_application(void) {
  _image->loadImage(_bbuf);

  if (!QMessageBox::warning(this, "Peinture Miroir " VERSION,
                            "Quit this application?\n\n",
                            "Quit", "Cancel"))
    close();
}

void PMiroir::setup_menu(void) {
  QMenuBar *menubar = menuBar();
  
  QPopupMenu *pmiro_menu = new QPopupMenu(this, "PMiro");
  pmiro_menu->insertItem(QString("Image Dir..."), this,
                         SLOT(setImagedir()));
  pmiro_menu->insertSeparator();
  pmiro_menu->insertItem(QString("Quit"), this,
                         SLOT(quit_application()));
  menubar->insertItem(QString("PMiro"), pmiro_menu);

  menubar->show();
}

void PMiroir::keyPressEvent(QKeyEvent *event) {
  setFullScreen();

  switch (event->key()) {
  case Qt::Key_Down:
  case Qt::Key_Left:
    if (_imagedir.isNull()) setImagedir();

    _mode = ImageMode;
    _image->loadPrevImage(_bbuf);
    updateRect(_fullrect);
    write_buf();
    break;
  case Qt::Key_Up:
  case Qt::Key_Right:
    if (_imagedir.isNull()) setImagedir();

    _mode = ImageMode;
    _image->loadNextImage(_bbuf);
    updateRect(_fullrect);
    write_buf();
    break;
  case Qt::Key_Return:
  case Qt::Key_F33:	// OK key
  case 0x2010:		// "*" beside LCD
    if (_mode == MirrorMode) {
      if (_imagedir.isNull()) setImagedir();

      _image->loadImage(_bbuf);
      _mode = ImageMode;
      updateRect(_fullrect);
      write_buf();
    } else {
      _mode = MirrorMode;
      resize(200, 200);
      repaint();
    }
    break;
  case Qt::Key_Escape:
    change_pen();
    break;
  case Qt::Key_1:
    _pen = &_pen1;
    break;
  case Qt::Key_2:
    _pen = &_pen2;
    break;
  case Qt::Key_3:
    _pen = &_pen3;
    break;
  default:
    event->ignore();
  }
}

void PMiroir::read_buf(void) {
  static const int w = Width;
  static const int h = Height;

  int x, y, offset, foffset;
  unsigned short int *fbuf, a;
  unsigned long int c;
  QDirectPainter dp(this);

  fbuf = (unsigned short int*)dp.frameBuffer();

  for (y = 0; y < h; y ++) {
    offset = y * w;
    foffset = h - 1 -y;
    c = 0;
    for (x = 0; x < w; x ++) {
      a = fbuf[foffset];
      _buf[offset ++] = a;
      c += a ^ x;
      foffset += h;
    }
    _chk[y] = c;
  }
}

void PMiroir::write_buf(void) {
  QRect r(0, 0, Width, Height);

  write_buf(r);
}

void PMiroir::write_buf(unsigned short int *b, QRect rect) {
  static copy_info_t cpi;

  cpi.x = rect.x();
  cpi.y = rect.y();
  cpi.w = rect.width();
  cpi.h = rect.height();
  cpi.bpp = 16;
  cpi.stride = Width * 16 / 8;
  cpi.adr = (unsigned char*)(b + (rect.y() * Width + rect.x()));
  ioctl(_vgafp, CFXGA_COPY_FB, &cpi);
}

void PMiroir::write_buf_dp(unsigned short int *b, QRect rect) {
  int x, y, x0, y0, x1, y1;

  x0 = rect.left();
  x1 = rect.right();
  y0 = rect.top();
  y1 = rect.bottom();

  QDirectPainter dp(this);
  unsigned short int *fbuf = 
    (unsigned short int *)dp.frameBuffer();

  for (y = y0; y <= y1; y ++)
    for (x = x0; x <= x1; x ++)
      fbuf[x * Height + (Height - 1 - y)] = b[y * Width + x];
}
  

void PMiroir::write_buf(QRect &rect) {
  write_buf(_buf, rect);
  if (_mode == ImageMode)
    write_buf_dp(_buf, rect);
}

void PMiroir::reload_buf(void) {
  if (_mode == MirrorMode) {
    exchange_buf();
    read_buf();
    update_buf();
    if (!menuBar()->isVisible())
      menuBar()->show();
  } else {
    if (menuBar()->isVisible())
      menuBar()->hide();
  }

  _timer->start(100, true);
}

void PMiroir::exchange_buf(void) {
  if (_buf == _buffer0) {
    _buf = _buffer1;
    _bbuf = _buffer0;
    _chk = _check1;
  } else {
    _buf = _buffer0;
    _bbuf = _buffer1;
    _chk = _check0;
  }
}

void PMiroir::update_buf(void) {
  int x0, x1, y, y0, y1;
  QRect rect;

  x0 = x1 = -1;
  for (y = 0; y < Height; y ++) {
    if (update_xpos(y, &x0, &x1)) {
      y0 = y;
      for (y ++; (y < Height) && (!update_xpos(y, &x0, &x1)); y ++) ;
      y1 = -- y;
      rect.setCoords(x0 * 2, y0, x1 * 2 + 1, y1);
      write_buf(rect);
      x0 = x1 = -1;
    }
  }
}

bool PMiroir::update_xpos(int y, int *x0, int *x1) {
  int offset = y * (Width / 2);
  int x, nx0, nx1;
  unsigned long int *buf32 = (unsigned long int*)_buf;
  unsigned long int *bbuf32 = (unsigned long int*)_bbuf;

  if ((_check0[y] == _check1[y]) || *x0 < 0) {
    for (x = 0;
         (buf32[offset + x] == bbuf32[offset + x]) && (x < Width / 2); x ++) ;
    if (x >= Width / 2) return false;
    nx0 = x;

    for (x = Width / 2 - 1;
         (buf32[offset + x] == bbuf32[offset + x]) && (x >= 0); x --) ;
    nx1 = x;
  } else {
    for (x = 0;
         (buf32[offset + x] == bbuf32[offset + x]) && (x < *x0); x ++) ;
    nx0 = x;

    for (x = Width / 2 - 1;
         (buf32[offset + x] == bbuf32[offset + x]) && (x > *x1); x --) ;
    nx1 = x;
  }

  if (*x0 < 0) *x0 = Width / 2;
  if (nx0 < *x0) *x0 = nx0;
  if (nx1 > *x1) *x1 = nx1;
  return true;
}

void PMiroir::convert_geometry(int *x, int *y) {
  if (qApp->desktop()->width() == Height) {
    int xx = *x;
    *x = *y;
    *y = Height - 1 - xx;
  }
}

void PMiroir::mousePressEvent(QMouseEvent *e) {
  if (_mode != ImageMode) {
    QMainWindow::mouseMoveEvent(e);
    return;
  }

  _x1 = e->globalX();
  _y1 = e->globalY();
  convert_geometry(&_x1, &_y1);

  _x1 = max(0, min(_x1, _painter->get_width() - 1));
  _y1 = max(0, min(_y1, _painter->get_height() - 1));

  _x0 = _x1;
  _y0 = _y1;
  
  mouseMoveEvent(e);
}

void PMiroir::mouseMoveEvent(QMouseEvent *e) {
  if (_mode != ImageMode) {
    QMainWindow::mouseMoveEvent(e);
    return;
  }

  _x0 = _x1; _y0 = _y1;
  
  _x1 = e->globalX();
  _y1 = e->globalY();
  convert_geometry(&_x1, &_y1);

  _x1 = max(0, min(_x1, _painter->get_width() - 1));
  _y1 = max(0, min(_y1, _painter->get_height() - 1));

  _painter->setPen(*_pen);
  _painter->drawLine(_x0, _y0, _x1, _y1);

  int min_x, min_y, max_x, max_y, pwz;
  pwz = (_pen->width() >> 1) + 1;
  if (_x0 <= _x1) {
    min_x = _x0 - pwz;
    max_x = _x1 + pwz;
  } else {
    min_x = _x1 - pwz;
    max_x = _x0 + pwz;
  }
  if (_y0 <= _y1) {
    min_y = _y0 - pwz;
    max_y = _y1 + pwz;
  } else {
    min_y = _y1 - pwz;
    max_y = _y0 + pwz;
  }


  QRect r = QRect(min_x, min_y,
                  max_x - min_x + 1,
                  max_y - min_y + 1) & 
    QRect(0, 0, Width, Height);
  
  updateRect(r);
  write_buf(r);
}

void PMiroir::mouseReleaseEvent(QMouseEvent *e) {
  if (_mode != ImageMode) {
    QMainWindow::mouseMoveEvent(e);
    return;
  }

  _x0 = _x1 = _y0 = _y1 = -1;

  _painter->clearMask();
  _painter->setPen(*_pen);

  setFullScreen();
}

void PMiroir::setFullScreen(void) {
  if (size() != qApp->desktop()->size()) {
    showNormal();
    setFixedSize(QApplication::desktop()->size());
    reparent(0, WType_TopLevel | WStyle_Customize |
             WStyle_NoBorderEx, QPoint(0, 0));
    showFullScreen();
  }
}

void PMiroir::resizeEvent(QResizeEvent *e) {
  if (_mode == ImageMode && size() == qApp->desktop()->size())
    write_buf_dp(_buf, QRect(0, 0, Width, Height));
  else
    QMainWindow::resizeEvent(e);
}

void PMiroir::paintEvent(QPaintEvent *e) {
  if (_mode == ImageMode && size() == qApp->desktop()->size()) {
    write_buf_dp(_buf, QRect(0, 0, Width, Height));
  } else {
    if (_mode == MirrorMode&& size() != QSize(200, 200)) {
      resize(200, 200);
    }
    QMainWindow::paintEvent(e);
  }
}

void PMiroir::updateRect(QRect r) {
  uint32 *buf0, *buf1, *this_buf;
  int x0, y0, x1, y1, w, w2;

  buf0 = (uint32*)_bbuf;
  buf1 = (uint32*)_painter->frameBuffer();
  this_buf = (uint32*)_buf;

  x0 = r.left();
  y0 = r.top();
  x1 = r.right();
  y1 = r.bottom();

  if (x0 & 1) x0 --;
  if (x1 & 1 != 1) x1 ++;
  w2 = x1 - x0 + 1;
  w = w2 >> 1;
  if (w2 & 1) w ++;


  int y, index, i;
  unsigned long int b, c, t, n;
  for (y = y0; y <= y1; y ++) {
    index = (y * Width + x0) >> 1;
    for (i = 0; i < w; i ++, index ++) {
      b = buf0[index];
      c = buf1[index];
      t = ((((b >> 17) & 0x7c00) | ((b >> 11) & 0x001f)) + 0x0401)
        * (((c >> 16) & 0xf800) | ((c >> 11) & 0x001f));
      n = ((t << 1) & 0xf8000000) | ((t << 6) & 0x0000f800);
      t = ((((b >> 12) & 0x7c00) | ((b >> 6) & 0x001f)) + 0x0401)
        * (((c >> 11) & 0xf800) | ((c >> 6) & 0x001f));
      n |= ((t >> 4) & 0x07c00000) | ((t << 1) & 0x000007c0);
      t = ((((b >> 6) & 0x7c00) | (b & 0x001f)) + 0x0401)
        * (((c >> 5) & 0xf800) | (c & 0x001f));
      n |= ((t >> 10) & 0x001f0000) | ((t >> 5) & 0x0000001f);
      this_buf[index] = n;
    }
  }
}

void PMiroir::change_pen(void) {
  repaint();

  QPainter p(this);
  p.setBackgroundColor(Qt::white);

  QString str;

  if (_pen == &_pen1) {
    _pen = &_pen2;
    str = "yellow pen";
  } else if (_pen == &_pen2) {
    _pen = &_pen3;
    str = "white pen";
  } else {
    _pen = &_pen1;
    str = "red pen";
  }

  p.setPen(Qt::white);
  p.drawText(100, 30, str);
  p.setPen(Qt::black);
  p.drawText(101, 31, str);
}

void PMiroir::setImagedir(void) {
  FileDialog dlg(this, "image dir", true, _imagedir);

  if (dlg.exec()) {
    QString fn = dlg.filePath();
    
    QFileInfo info(fn);
    _imagedir = info.dirPath();
    _image->setupDir(_imagedir);
  }
}
